//#define USE_INTERNAL_GRF_WAIT
//#define USE_INTERNAL_MSG_WAIT


#include <windows.h>
#include <streams.h>
#include <shlwapi.h>
/// this is required by the BaseClasses
#pragma comment(lib,"winmm")
/// if you do not want a dependency on ShLWAPI.DLL, you need to provide functions equivalent to wnsprintf() and StrStrI()
#pragma comment(lib,"shlwapi")


#define REFTIME_UNITS_PER_MILLISEC	10000i64
#define REFTIME_UNITS_PER_SEC		10000000i64


/// just a useful utility function
template <class C, class I> HRESULT STDMETHODCALLTYPE CreateInstance(I** ppv)
{
	if(ppv == NULL) return E_POINTER; else *ppv = NULL;
	HRESULT hr = S_OK;
	C* p = new C(&hr);
	if(NULL == p) { hr = E_OUTOFMEMORY; goto exit; }
	if(S_OK != hr) goto exit;
	if(S_OK != (hr = p->QueryInterface(__uuidof(I),(void**)ppv))) goto exit;
	p = NULL;
exit:
	if(p) delete p;
	return hr;
}


class CSomeDevelopersMustDieGraphBuilderCallback : public CUnknown, public IAMGraphBuilderCallback
{
friend HRESULT STDMETHODCALLTYPE CreateInstance<CSomeDevelopersMustDieGraphBuilderCallback,IAMGraphBuilderCallback>(IAMGraphBuilderCallback** ppv);
private:
	CSomeDevelopersMustDieGraphBuilderCallback(HRESULT* phr) : CUnknown(TEXT("CSomeDevelopersMustDieGraphBuilderCallback"),NULL,phr) {}

	/// *** INonDelegatingUnknown *** ///
protected:
	HRESULT STDMETHODCALLTYPE NonDelegatingQueryInterface(REFIID riid, void** ppv)
	{
		if(riid == __uuidof(IAMGraphBuilderCallback)) return GetInterface((IAMGraphBuilderCallback*)this,ppv);
		return CUnknown::NonDelegatingQueryInterface(riid,ppv);
	}

	/// *** IUnknown *** ///
public:
	DECLARE_IUNKNOWN

	/// *** IAMGraphBuilderCallback *** ///
public:
	HRESULT STDMETHODCALLTYPE SelectedFilter(IMoniker* pMoniker)
	{
		HRESULT			hr		= S_OK;
		DWORD			rc		= ERROR_SUCCESS;
		IPropertyBag*	pBag	= NULL;
		HKEY			hKey	= NULL;
		VARIANT			v;
		TCHAR			s[MAX_PATH + 1];
		DWORD			t, n;
		CLSID			clsid;

		VariantInit(&v);

		/// get the CLSID of the referenced filter
		if(S_OK != (hr = pMoniker->BindToStorage(NULL,NULL,__uuidof(IPropertyBag),(void**)&pBag))) goto exit;
		if(S_OK != (hr = pBag->Read(L"CLSID",&v,NULL))) goto exit; /// this is undocumented
		if(v.vt != VT_BSTR || v.bstrVal == NULL) { hr = E_UNEXPECTED; goto exit; }
		if(S_OK != (hr = CLSIDFromString(v.bstrVal,&clsid))) goto exit;
		(void)VariantClear(&v);

		/// some VCM aka ICM aka VFW decoders crash when released
		if(clsid == CLSID_AVIDec) { hr = E_FAIL; goto exit; }
		/// this filter crashes too easily even when not used
		if(clsid == CLSID_AVIDraw) { hr = E_FAIL; goto exit; }

		/// get the filter's DLL path
		if(wnsprintf(s,sizeof(s)/sizeof(*s) - 1,TEXT("CLSID\\%ls\\InprocServer32"),v.bstrVal) < 0) { hr = E_UNEXPECTED; goto exit; }
		if(ERROR_SUCCESS != (rc = (DWORD)RegOpenKeyEx(HKEY_CLASSES_ROOT,s,0,KEY_QUERY_VALUE,&hKey))) { hr = HRESULT_FROM_WIN32(rc); goto exit; }
		if(ERROR_SUCCESS != (rc = (DWORD)RegQueryValueEx(hKey,NULL,NULL,&t,(LPBYTE)s,((n = sizeof(s)/sizeof(*s) - 1),&n)))) { hr = HRESULT_FROM_WIN32(rc); goto exit; }
		if(t != REG_SZ) { hr = E_UNEXPECTED; goto exit; } else s[sizeof(s)/sizeof(*s) - 1] = TEXT('\0');

		/// this avoids all third-party filters, but not third-party DMOs
		if(NULL == StrStrI(s,TEXT("quartz.dll"))) { hr = E_FAIL; goto exit; }

	exit:
		(void)VariantClear(&v);
		return hr;
	}
	HRESULT STDMETHODCALLTYPE CreatedFilter(IBaseFilter* pFilter)
	{
		return S_OK;
	}
};


/// IMediaEvent::WaitForCompletion() would be enough in this case, but this is an excuse
/// to provide a sample of how to keep dispatching messages while waiting for a graph event
/// in way similar to how the graph's own IMediaEvent::WaitForCompletion() works
HRESULT STDAPICALLTYPE WaitForCompletion(IUnknown* pGraph, DWORD dwTimeout, long* plEvent, HRESULT* phr)
{
	HRESULT			hr		= S_OK;
	IMediaEvent*	pEvent	= NULL;
#ifdef USE_INTERNAL_GRF_WAIT
	HANDLE			hEvent;
#endif

	if(pGraph == NULL) { hr = E_INVALIDARG; goto exit; }

	if(plEvent) *plEvent = 0;
	if(phr) *phr = S_OK;

	if(S_OK != (hr = pGraph->QueryInterface(__uuidof(IMediaEvent),(void**)&pEvent))) goto exit;

#ifdef USE_INTERNAL_GRF_WAIT
	if(S_OK != (hr = pEvent->GetEventHandle((OAEVENT*)&hEvent))) goto exit;

	for(DWORD t = timeGetTime();;) {
		switch(DWORD rc = 
#ifdef USE_INTERNAL_MSG_WAIT
			MsgWaitForMultipleObjects(1,&hEvent,FALSE,dwTimeout,QS_ALLINPUT)
#else
			WaitDispatchingMessages(hEvent,dwTimeout,NULL,0)
#endif
		)
		{
			case WAIT_FAILED:
				hr = HRESULT_FROM_WIN32(GetLastError()); goto exit;
			case WAIT_TIMEOUT:
				hr = E_ABORT; goto exit;
#ifdef USE_INTERNAL_MSG_WAIT
			case WAIT_OBJECT_0 + 1: {
				for(MSG msg; PeekMessage(&msg,NULL,0,0,PM_REMOVE);) (void)DispatchMessage(&msg);
				if(dwTimeout != INFINITE) { DWORD d = timeGetTime() - t; if(d < dwTimeout) dwTimeout -= d; else { hr = E_ABORT; goto exit; } }
			} break;
#endif
			case WAIT_OBJECT_0:
			case WAIT_ABANDONED_0: {
				long lEvent; LONG_PTR lParam1, lParam2;
				if(S_OK != (hr = pEvent->GetEvent(&lEvent,&lParam1,&lParam2,0))) goto exit;
				(void)pEvent->FreeEventParams(lEvent,lParam1,lParam2);
				switch(lEvent)
				{
					case EC_ERRORABORT:
						if(phr) *phr = (HRESULT)lParam1;
					case EC_COMPLETE:
					case EC_USERABORT:
						if(plEvent) *plEvent = lEvent; goto exit;
				}
			} break;
			default:
				hr = E_UNEXPECTED; goto exit;
		}
	}
#else /// USE_INTERNAL_GRF_WAIT
	{
		long lEvent;
		if(S_OK != (hr = pEvent->WaitForCompletion(dwTimeout,&lEvent))) goto exit;
		if(plEvent) *plEvent = lEvent;
	}
#endif /// USE_INTERNAL_GRF_WAIT

exit:
	if(pEvent) pEvent->Release();
	return hr;
}


/// riid: must support the given interface, unless riid == GUID_NULL
/// REGUID majortype, ...: a sequence of acceptable media types, terminated by MEDIATYPE_NULL
/// VFW_E_NOT_FOUND if not found
HRESULT STDAPICALLTYPE FindFirstPin(IUnknown* pFilter, bool connected, PIN_DIRECTION dir, REFIID riid, IPin** ppPin, REFGUID majortype, ...)
{
	HRESULT				hr		= S_OK;
	IBaseFilter*		pBase	= NULL;
	IEnumPins*			pPins	= NULL;
	IPin*				pPin	= NULL;
	IPin*				pFound	= NULL;
	IEnumMediaTypes*	pTypes	= NULL;
	AM_MEDIA_TYPE*		pmt		= NULL;

	if(ppPin == NULL) { hr = E_POINTER; goto exit; } else *ppPin = NULL;
	if(pFilter == NULL) { hr = E_INVALIDARG; goto exit; }

	if(S_OK != (hr = pFilter->QueryInterface(__uuidof(IBaseFilter),(void**)&pBase))) goto exit;

	if(S_OK != (hr = pBase->EnumPins(&pPins))) goto exit;
	while(pFound == NULL && S_OK == (hr = pPins->Next(1,&pPin,NULL))) {
		PIN_DIRECTION d; IPin* p;
		switch(hr = pPin->ConnectedTo(&p))
		{
			case S_OK: p->Release(); if(connected) break; else goto next;
			case VFW_E_NOT_CONNECTED: if(connected) goto next; else break;
			default: goto exit;
		}
		if(S_OK != (hr = pPin->QueryDirection(&d))) goto exit; else if(d != dir) goto next;
		if(majortype == MEDIATYPE_NULL) {
			pFound = pPin;
		} else {
			for(LPCGUID t = &majortype; pFound == NULL && *t != MEDIATYPE_NULL; t++) {
				if(S_OK != (hr = pPin->EnumMediaTypes(&pTypes))) goto exit;
				while(pFound == NULL && S_OK == (hr = pTypes->Next(1,&pmt,NULL))) {
					if(pmt->majortype == *t) pFound = pPin;
					DeleteMediaType(pmt); pmt = NULL;
				} if(pFound == NULL && S_OK != hr && S_FALSE != hr) goto exit; else hr = S_OK;
				pTypes->Release(); pTypes = NULL;
			}
		}
		if(pFound) pFound->AddRef();
	next:
		pPin->Release(); pPin = NULL;
	} if(pFound == NULL && S_OK != hr && S_FALSE != hr) goto exit; else hr = S_OK;

	if(pFound) { *ppPin = pFound; pFound = NULL; } else hr = VFW_E_NOT_FOUND;

exit:
	if(pmt		) DeleteMediaType(pmt);
	if(pTypes	) pTypes	->Release();
	if(pFound	) pFound	->Release();
	if(pPin		) pPin		->Release();
	if(pPins	) pPins		->Release();
	if(pBase	) pBase		->Release();
	return hr ;
}

HRESULT STDAPICALLTYPE AddGraphToRot(IUnknown* pGraph, DWORD* pdwCookie) 
{
	HRESULT					hr			= S_OK;
	IMoniker*				pMoniker	= NULL;
	IRunningObjectTable*	pTable		= NULL;
	WCHAR					szName[64];

	if(S_OK != (hr = GetRunningObjectTable(0,&pTable))) goto exit;
	(void)wnsprintfW(szName,sizeof(szName)/sizeof(*szName),L"FilterGraph %08lX pid %08lX",pGraph,GetCurrentProcessId());
	if(S_OK != (hr = CreateItemMoniker(L"!",szName,&pMoniker))) goto exit;
	if(S_OK != (hr = pTable->Register(ROTFLAGS_REGISTRATIONKEEPSALIVE,pGraph,pMoniker,pdwCookie))) goto exit;

exit:
	if(pMoniker	) pMoniker	->Release();
	if(pTable	) pTable	->Release();
	return hr;
}

HRESULT STDAPICALLTYPE RemoveGraphFromRot(DWORD dwCookie) 
{
	HRESULT					hr			= S_OK;
	IRunningObjectTable*	pTable		= NULL;

	if(dwCookie == 0) goto exit;
	if(S_OK != (hr = GetRunningObjectTable(0,&pTable))) goto exit;
	if(S_OK != (hr = pTable->Revoke(dwCookie))) goto exit;

exit:
	if(pTable	) pTable	->Release();
	return hr;
}

HRESULT STDAPICALLTYPE SaveGraphToGrfFile(IUnknown* pGraph, LPCWSTR szFile) 
{
const WCHAR wszStreamName[] = L"ActiveMovieGraph"; 
	HRESULT			hr			= S_OK;
	IStorage*		pStorage	= NULL;
	IStream*		pStream		= NULL;
	IPersistStream*	pPersist	= NULL;

	if(S_OK != (hr = StgCreateDocfile(szFile,STGM_CREATE|STGM_TRANSACTED|STGM_READWRITE|STGM_SHARE_EXCLUSIVE,0,&pStorage))) goto exit;
	if(S_OK != (hr = pStorage->CreateStream(L"ActiveMovieGraph",STGM_WRITE|STGM_CREATE|STGM_SHARE_EXCLUSIVE,0,0,&pStream))) goto exit;
	if(S_OK != (hr = pGraph->QueryInterface(__uuidof(IPersistStream),(void**)&pPersist))) goto exit;
	if(S_OK != (hr = pPersist->Save(pStream,TRUE))) goto exit;
	if(S_OK != (hr = pStorage->Commit(STGC_DEFAULT))) goto exit;

exit:
	if(pPersist	) pPersist	->Release();
	if(pStream	) pStream	->Release();
	if(pStorage	) pStorage	->Release();
	return hr;
}

EXTERN_C void WINAPI SetThreadName(LPCSTR szThreadName)
{
#pragma pack(4)
	struct {
		DWORD	dwType;		// Must be 0x1000.
		LPCSTR	szName;		// Pointer to name (in user addr space).
		DWORD	dwThreadID;	// Thread ID (-1=caller thread).
		DWORD	dwFlags;	// Reserved for future use, must be zero.
	} info = { 0x1000, szThreadName, GetCurrentThreadId(), 0 };
#pragma pack()
	__try { RaiseException(0x406D1388,0,sizeof(info)/sizeof(DWORD),(DWORD*)&info); } __except(EXCEPTION_CONTINUE_EXECUTION) {}
}
