#ifndef _WmvMux_H
#define _WmvMux_H


#include <windows.h>
#include <wmsdkidl.h>


#define NAME_AsfMuxFilter L"AsfMuxFilter"

class __declspec(uuid("{DD2EFBC3-DEB2-4d3f-9AAF-16CA65F1324C}")) CAsfMuxFilter;
#define CLSID_AsfMuxFilter __uuidof(CAsfMuxFilter)


class __declspec(uuid("{59856708-6220-482c-BA0F-D52A69E4E2DB}")) IAsfMuxFilter : public IUnknown
#define IID_IAsfMuxFilter __uuidof(IAsfMuxFilter)
{
public:
	/**
	 * Can be called at any time but has no effect after streaming stops. Defaults to false.
	 * @param enable if TRUE, the filter stops writing at the first end-of-stream, otherwise at the last one
	 */
	virtual void STDMETHODCALLTYPE StopAtShortestStream(BOOL enable) = 0;
	/**
	 * @param enable if TRUE, the filter spawns an output pin compatible with the FileWriter and refuses to add sinks
	 * @param szFile if not NULL, the output pin will expose IStreamBuilder and connect itself to an unconnected writer
	 *	             (that is, a filter that implements IFileSinkFilter and has 1 input pin that supports IStream)
	 *	             and configures it to write to szFile (or overwrite it, if the writer supports IFileSinkFilter2),
	 *	             unless the writer was already configured; if there is no unconnected writer in the graph, a new CLSID_FileWriter
	 *	             is created
	 * @return VFW_E_WRONG_STATE if enable is TRUE but there already are sinks or enable is FALSE but the output pin is already connected
	 */
	virtual HRESULT STDMETHODCALLTYPE EnableAsfMuxer(BOOL enable, LPCWSTR szFile) = 0;
	/**
	 * @param pSink any object that implements IWMWriterSink
	 * @return VFW_E_WRONG_STATE if muxing is enabled
	 */
	virtual HRESULT STDMETHODCALLTYPE AddAsfSink(IUnknown* pSink) = 0;
	/**
	 * @param zDoAutoIndexing see IWMWriterFileSink3::SetAutoIndexing()
	 * @param ppSink can be NULL
	 * @return VFW_E_WRONG_STATE if muxing is enabled
	 */
	virtual HRESULT STDMETHODCALLTYPE AddAsfFileSink(LPCWSTR szFile, BOOL zDoAutoIndexing, IWMWriterFileSink3** ppSink) = 0;
	/**
	 * @param wPort 0 for system selection
	 * @param pwPort can be NULL; same as wPort if wPort != 0
	 * @param ppSink can be NULL
	 * @return VFW_E_WRONG_STATE if muxing is enabled
	 */
	virtual HRESULT STDMETHODCALLTYPE AddAsfNetworkSink(WORD wPort, WORD* pwPort, IWMWriterNetworkSink** ppSink) = 0;
	/**
	 * @param ppSink can be NULL
	 * @return VFW_E_WRONG_STATE if muxing is enabled
	 */
	virtual HRESULT STDMETHODCALLTYPE AddAsfStreamSink(IStream* pStream, IWMWriterSink** ppSink) = 0;
};


/**
 * Same as ::CoCreateInstance(CLSID_AsfMuxFilter,pOuter,CLSCTX_INPROC_SERVER,riid,ppv).
 */
EXTERN_C HRESULT STDAPICALLTYPE CreateAsfMuxFilter(IUnknown* pOuter, REFIID riid, void** ppv);


#endif /// _WmvMux_H
